/*
  Class:  Disk

  Author: Alexander Rasin

  This is a class that keeps the player table (in hash table) that is simply
  a mapping between a user name and a provile.
  This class can add new profiles to the table, and look them up.
*/

package scorchserver;

import java.io.*;
import java.util.*;

import scorch.*;
import scorch.utility.Crypt;

public class Disk implements Runnable
{
    private BufferedReader in;
    private BufferedWriter out;
    private String st, fileName;
    private static final String table = "players.db";
    private Date lastBackup;
    private Hashtable hash = null;
    private Thread thread = null;
    //the time to wait between checking whether profiles need to be
    //written to disk.
    private int WRITE_DELAY = 15*60*1000;
    private boolean changed = false, keep_running = true;

    private static final String desyncLogFile = "desync.";
    private static int desyncCount = 0;

    //constructor
    public Disk()
    {
	File f = new File(table);

	lastBackup = new Date();	

	try {
	    //make a new file for profile storage if it does not exist
	    if (!f.exists())
		{
                    System.out.println("Creating new "+table+" file\n");
                    out = new BufferedWriter(new FileWriter(table));
                    out.close();
                    out = null;
		}
	    //	f.createNewFile();

	    // load the table from file to memory (hashtable)
	    in = new BufferedReader(new FileReader(table));
	    loadTable();
	    in.close();
	    in = null;
	    changed = false;

	    thread = new Thread( this );
	    thread.start();
	}
	catch(Exception e) {
	    System.err.println("Warning: Unable to read/write profiles " + e);
	}

    }

    public int getRegisteredUsersCount()
    {
	return hash.size();
    }
    
    //add a new profile.
    //currently used to replace a profile
    public synchronized void add(PlayerProfile profile)
    {
	if (hash == null) return;

	if (profile != null && hash.get(profile.getName()) != null)
	   System.out.println("WARNING: profile for " + profile.getName() + " already exists (overwriting)");

	changed = true;
	ScorchServer.insertTopTen(profile);
	hash.put(profile.getName(), profile);
    }

    public static synchronized void recordGame(Game g, String os, String ns)
    {
	BufferedOutputStream outBuf = null;

	String temp = os + "\n" + ns + "\nSeed: " + g.getSeed() + "\n" + g
	    + g.getAllHumanPlayersString();
	
	try
	    {
		outBuf = new BufferedOutputStream
		    (new FileOutputStream(Disk.desyncLogFile
					  +Disk.desyncCount++));

		for (int i = 0; i < temp.length(); i++)
		    outBuf.write(temp.charAt(i));

		outBuf.close();
	    }
	catch (IOException e)
	    {
		System.err.println("Disk: Desync Log write failed " + e);
	    }
    }

    public static synchronized void recordClientLog(String log, int cl_id)
    {
	BufferedOutputStream outBuf = null;

	try
	    {
		outBuf = new BufferedOutputStream
		    (new FileOutputStream("desync." + (Disk.desyncCount-1) +
					  "." + cl_id));

		log = log.replace(Protocol.separator, '\n');
		
		for (int i = 0; i < log.length(); i++)
		    outBuf.write(log.charAt(i));

		outBuf.close();
	    }
	catch (IOException e)
	    {
		System.err.println("Disk: Client Log write failed " + e);
	    }
    }

    //calls add so don't need to change the boolean flag changed
    public synchronized void change(PlayerProfile newProfile)
    {
	String name = newProfile.getName();

	remove(name);
	add(newProfile);
    }

    /*public synchronized PlayerProfile findProfileByName( String name )
    {
	if (hash == null) return null;
	Enumeration e = hash.elements();
	PlayerProfile prof = null;

	while( e.hasMoreElements() )
	    {
		prof = (PlayerProfile)e.nextElement();
		if (prof.getName().equals(name))
		    return prof;
	    }

	return null;
	}*/

    public synchronized void encryptPasswords()
    {
	if (hash == null) return;
	Enumeration e = hash.elements();

	while( e.hasMoreElements() )
	    add(((PlayerProfile)e.nextElement()).encrypt());
    }

    public static void main(String args[])
    {
	
	try
	    {
		BufferedReader bin = new BufferedReader(new FileReader(Disk.table));
		Disk.loadTableFiltered(bin);
		bin.close();
		bin = null;
	    }
	catch(IOException e)
	    {
		System.err.println(e);
	    }
	
	//Disk disk = new Disk();
	//disk.encryptPasswords();

	//	System.out.println("a is " + (byte)'a' + " A is " + (byte)'A' + " z is " 
	//		   + (byte)'z' + " Z is " + (byte)'Z');
    //System.out.println(" 0 "+(byte)'0' + " 9 is " + (byte)'9');
	//disk.add(new PlayerProfile("ME2342", "Parol' "));
	//disk.add(new PlayerProfile("ME1344", "Parol' "));
	//disk.add(new PlayerProfile("shuri165234k", "Parol' "));
	
	//String blah = new String("command bl12 23 152");
	
	//System.out.println(blah.substring("command".length()+1, blah.length()));
	//disk.writeTable();
    }

    public PlayerProfile getProfile( String user )
    {
	Object pr;

	if (user == null || hash == null)
	    return null;

	pr = hash.get(user);

	if (pr != null)
	    {
		long now = (new Date()).getTime();
		((PlayerProfile)pr).setAccessTime(now);

		return (PlayerProfile)pr;
	    }

	return null;
    }

    public void remove(String name)
    {
	if (hash != null)
	    hash.remove(name);
    }

    public static int getDesyncCount()
    {
	return Disk.desyncCount;
    }

    public void loadTable()
    {
	String profileString = null;
	PlayerProfile profile = null;
	long time = new Date().getTime();
	int count = 0;

	hash = new Hashtable();
	
	try {
	    //read the profiles from the profile file.
	    while ((profileString = in.readLine()) != null)
		{
		    profile = new PlayerProfile(profileString);
		    add(profile);
		    //sorted insert, will only work if score is in top ten
		    ScorchServer.insertTopTen(profile);
		    count++;
		}
	}
	catch (Exception e) 
	    {
		System.err.println("table load failed: " + e);
	    } 
	System.out.println(" Profile table of (" + count + ") loaded in " +
			   ((new Date().getTime() - time)/1000.0) +
			   " seconds ");
    }
    
    public static void loadTableFiltered(BufferedReader in)
    {
	String profileString = null;
	PlayerProfile profile = null;
	long time = new Date().getTime();
	int count = 0, countFiltered = 0;

	Hashtable hash1 = new Hashtable();
	Hashtable hash2 = new Hashtable();

	try {
	    //read the profiles from the profile file.
	    while ((profileString = in.readLine()) != null)
		{
		    profile = new PlayerProfile(profileString);
		    if( 			
		       //(profile.getName().length() > 10 ||
		       //profile.getName().length() < 3) &&
		       profile.getEmail().indexOf('@') == -1 &&
			profile.getOverallKills() == 0 &&
			profile.getOverallGain() == 0 &&
			profile.getTankType() == -1 &&
			profile.getCrationTime() == 
			profile.getAccessTime() )
			{
			    System.out.println(profile.getName()+" | "+profile.getEmail());
			    hash2.put(profile.getName(), profile);
			    countFiltered++;
			}
		    else
			{
			    hash1.put(profile.getName(), profile);
			    count++;
			}
		}
	}
	catch (Exception e) 
	    {
		System.err.println("table load failed: " + e);
	    } 
	System.out.println(" Profile table of (" + count + ") loaded in " +
			   ((new Date().getTime() - time)/1000.0) +
			   " seconds. Filtered: "+countFiltered);

	// now save both files
	try
	    {
		Enumeration hashElts = hash1.elements();
		String elt;
		BufferedWriter bout = new BufferedWriter(new OutputStreamWriter
		    (new FileOutputStream("players-good.db")));
		
		while (hashElts.hasMoreElements()) 
		    {
			try
			    {
				elt = ((PlayerProfile)hashElts.nextElement()).toString();
				bout.write(elt, 0, elt.length());
				bout.write('\n');
			    }
			catch( Throwable t )
			    {
				// weird problem in Kaffe only (??)
				System.err.println("failed to write profile: "+t);
			    }
		    }
		bout.flush();
		bout.close();
                bout = null;		

		hashElts = hash2.elements();
		bout = new BufferedWriter(new OutputStreamWriter
		    (new FileOutputStream("players-bad.db")));
		
		while (hashElts.hasMoreElements()) 
		    {
			try
			    {
				elt = ((PlayerProfile)hashElts.nextElement()).toString();
				bout.write(elt, 0, elt.length());
				bout.write('\n');
			    }
			catch( Throwable t )
			    {
				// weird problem in Kaffe only (??)
				System.err.println("failed to write profile: "+t);
			    }
		    }
		bout.flush();
		bout.close();
                bout = null;
	    }
	catch(IOException e)
	    {
		System.out.println("writing: "+e);
	    }
	
    }
    

    //public void writeTable()
    public void run()
    {
	Enumeration /*hashKeys = null,*/ hashElts = null;
	String elt = null;

	while( keep_running ) {
	    try 
		{
		    
		    while (!changed)
			{
			    Thread.sleep(WRITE_DELAY);
			}
		    
		    //synchronized (this) 
			{
			if (hash != null) {
			    //hashKeys = hash.keys();
			    hashElts = hash.elements();
			    
			    // create a new stream (and thus empty the file?)
			    // but do this after we get elements
			    // so now if we run out of memory we will
			    // preserve latest db
			    Date curDate = new Date();
			    if( (curDate.getTime() - lastBackup.getTime()) / 1000 > 3600*24 )
			    {	/* 24 hours since last backup */	
			   	out = new BufferedWriter(new OutputStreamWriter
					(new FileOutputStream(table+"-"+
					curDate.toString().replace(':','.'))));
				lastBackup = curDate;
			    }
			    else
			    {
			    	out = new BufferedWriter(new OutputStreamWriter
						(new FileOutputStream(table+"-new")));
			    }
 
			    long time = curDate.getTime();

			    int i = 0;
			    //while (hashKeys.hasMoreElements()) {
			    while (hashElts.hasMoreElements()) {
				try
				    {
					//elt = ((PlayerProfile)hash.get
					//       ((String)hashKeys.nextElement())) +"";
					elt = ((PlayerProfile)hashElts.nextElement()).toString();
					out.write(elt, 0, elt.length());
					out.write('\n');
				    }
				catch( OutOfMemoryError oome)
				    {
					System.err.println("failed to write profile, out of memory");
					System.gc();
					try { Thread.sleep(20*1000); } 
					catch(InterruptedException e){}
					System.err.println("aborted this db write");
					out.flush();
					out.close();
					out = null;
					return;
				    }
				catch( Throwable t )
				    {
					// weird problem in Kaffe only (??)
					System.err.println("failed to write profile: "+t);
				    }
				
				if( i++ % 5 == 0 )
				    Thread.sleep( 20 );
			    }
			    out.flush();
			    out.close();
                            out = null;
			    changed = false;
			    
			    System.out.println(" Table write completed in " +
					       ((new Date().getTime() - 
						 time)/1000.0) + " seconds ");
			    File src = new File(table+"-new");
			    File dst = new File(table);
			    boolean result;
                            result = dst.delete();
                            if (!result)
                            {
                               System.err.println(new Date()+" Failed to remove "+dst);
                            } 
                            result = src.renameTo(dst);
                            if (!result)
                            {
				System.err.println(new Date()+" Failed to rename "+table+"-new" + " to " + table + "; "+src+" exists: "+src.exists());
                            } 
			}
		    }
		}
	    catch (Throwable e) 
		{
		    out = null;
		    hashElts = null;
		    System.gc();
		    System.err.println("table write failed: " + e);
		    try
			{
			    out.flush();
			    out.close(); 
                            out = null;
			}
		    catch(IOException e1)
			{}
		}
	}
	System.out.println("Disk exiting... " );
	hash = null;
	in = null;
	out = null;
    }
    
    public void shutdown()
    {
	changed = true;
	keep_running = false;
    }
}
