/*
  Class:  ScorchServer

  Author: Alexander Rasin

  ScorchServer class is the class representing the actual server.  It's 
  purpose is to listen to client connections and once a connection is 
  established, it starts an instance of a ServerThread that will be 
  representing the client on the server side.
  ScorchServer also keeps track of the games in order to be able to
  run global checks for a user that has already in another game.  It
  also has access to class Disk that stores and looks up profiles - 
  games can call on the Server in order to lookup or write a profile of 
  a player.
  Once logged in succesfully, all the clients register on the server 
  that either assigns a client to existing game or makes a new one if 
  there are no spots available.
*/
  
package scorchserver;

import java.net.*;
import java.util.*;
import java.io.*;
import scorch.PlayerProfile;
import scorch.Protocol;
import scorchserver.servershell.*;

public class ScorchServer
{
    public static final String 
	version = "v1.36 12/10/2004",
	fullName = "Scorched Earth 2000 Server, "+version;

    private Socket accept = null;
    private ServerSocket s_socket = null;
    private boolean shut_down = false;
    private int port;
    private static int game_count = 0;
    private static Date up_time = null;
    private static int maxGameCount = 0, maxMGameCount = 0, maxUserCount = 0;
    public static int potentialDesyncCount = 0;
    
    //all the games on the server
    private static Vector games = new Vector();
    //responsible for profile storage and lookup
    private static Disk disk = null;
    private static Vector topTen = new Vector();
    //can modify top ten to top 'any other number'
    private static int topX = 10;
    private ServerStats serverStats;

    public static void main(String[] args)
    {
	try
	    {
		System.setErr(new PrintStream(new FileOutputStream("err.log"), true));
	    }
	catch(IOException e)
	    {
		System.err.println("error redirecting err: "+e);
	    }
	
	disk = new Disk();
	ScorchServer ss = null;
	ServerShell shell = null;

	System.out.println(fullName+"\n");
	System.out.println
	    ("Copyright (C) 1999-2004 KAOS Software\n"+
	     "Scorched Earth 2000 comes with ABSOLUTELY NO WARRANTY;\nThis is"+
	     " free software, and you are welcome to redistribute it\nunder"+
	     " certain conditions. Please read COPYING for details.\n");
	
	try 
	    {
		//new Thread(new RemoteServerShell()).start();
		up_time = new Date();
		
		if (args.length > 0)
		    ss = new ScorchServer(Integer.parseInt(args[0]));
		else
		    ss = new ScorchServer(4242);

		shell = new ServerShell( ss );
		new Thread( shell ).start();
		ss.runServer();
	    }
	catch (Throwable t)
	    {
		System.out.println("ERROR starting server: "+t);
	    }
	System.exit( 0 );
    }
    
    public ScorchServer(int port)
    {
	this.port=port;
	serverStats = new ServerStats(this);
	//runServer();
    }
 
    //check if a player is allready participating in some game
    public static synchronized boolean alreadyPlaying(String name)
    {
	Game g = null;

	//see if the player is already logged into one of the games
	for (int i = 0; i < games.size(); i++)
	    {
		g = (Game)games.elementAt(i);
		if (g.alreadyPlaying(name))
		    return true;
	    }
	return false;
    }

    //lookup a player in the profile hashtable
    public static synchronized PlayerProfile lookupPlayer(String name)
    {
	return disk.getProfile(name);
    }
    
    //write the hashtable to disk
    //    public static synchronized void writeTable()
    //{
    //	disk.writeTable();
    //}
    
    //makes a new player in the profile database. 
    public static synchronized void newPlayer(PlayerProfile profile)
    {
	long now = (new Date()).getTime();
        profile.setCrationTime(now);
        profile.setAccessTime(now);
	disk.add(profile);
	//writeTable();
    }
    
    //change user's profile (as requested by the user)
    public static synchronized void changeProfile(PlayerProfile profile)
    {
	//System.out.println("CHANGING " + profile);
	disk.change(profile);
	//System.out.println("RESULT " + lookupPlayer(profile.getName()));
	//writeTable();
    }
    
    //removes the user profile
    public static synchronized void deleteProfile(PlayerProfile profile)
    {
    
    }

    public static String allGamesToString(boolean verbose)
    {
	String all = "";
	Game g = null;
	int multi_player = 0;
	
	Enumeration enumGames = games.elements();
	while(enumGames.hasMoreElements())
	    {
		g = (Game)enumGames.nextElement();
		if (g.getHumanCount() > 1)  multi_player++;
		if( verbose ) all = all + g + "\n";
	    }
	
	if( verbose )
	    {
		all += "Total of " + games.size() + " games, " + multi_player +
		    " multiplayer.\n\n";
	    }
	else
	    {
		all = games.size() + " ("+multi_player+")";
	    }

	return all;
    }

    public static int getDesyncCount()
    {
	return Disk.getDesyncCount();
    }

    public static int getGameCount()
    {
	return game_count;
    }

    public static String getMaxGamesCount()
    {
	return maxUserCount + " users in " + maxGameCount+" ("+
	    maxMGameCount+") games";
    }

    public static int getPlayersCount()
    {
	Game g;
	int res = 0;

	for (int i = 0; i < games.size(); i++)
	    {
		g = (Game)games.elementAt(i);
		res += g.getHumanPlayersCount();
	    }
	return res;
    }

    public static String getUpTime()
    {
	Date date = new Date();
	long timed = date.getTime()-up_time.getTime();
	String uptime = "";
	timed /= 1000;
	uptime += (timed/(3600*24)) + " days ";
	timed %= (3600*24);
	uptime += (timed/(3600)) + " hours ";
	timed %= (3600);
	uptime += (timed/(60)) + " minutes ";
	timed %= (60);
	uptime += (timed) + " seconds";
	return uptime;
    }

    public static Game findGameByID(int index)
    {
	Game g;

	for (int i = 0; i < games.size(); i++)
	    {
		g = ((Game)games.elementAt(i));
		if (g.getID() == index)
		    return g;
	    }

	return null;
    }

    /*public static PlayerProfile findProfileByName( String name )
    {
	return disk.findProfileByName( name );
	}*/

    public static Player findPlayerByName( String name )
    {
	Player pl = null;

	for (int i = 0; i < games.size(); i++)
	    {
		pl = ((Game)games.elementAt(i)).findPlayerByName(name);
		if (pl != null)
		    return pl;
	    }

	return null;
    }

    public static void insertTopTen(PlayerProfile profile)
    {
	int position = 0;

	//the profile is already in the top list
	//	if (topTen.contains(profile))
	//    return;

	for (int i = 0; i < topTen.size(); i++)
	    if (profile.getName().equals
		(((PlayerProfile)topTen.elementAt(i)).getName()))
		topTen.removeElementAt(i);

	while (position<topX && position < topTen.size())
	    {
		if (profile.getOverallGain() > 
		    ((PlayerProfile)topTen.elementAt
		     (position)).getOverallGain())
		    break;
		
		position++;
	    }

	if (position == topTen.size() && !(position<topX) )
	    return;

	//System.out.println("Inserting " + profile.getName() + " at " + position);
	topTen.insertElementAt(profile, position);

	//if we are over the size of hall of fame, remove the last player
	if (topTen.size() > topX)
	    topTen.removeElement(topTen.lastElement());
    }

    public static String getTopTen()
    {
	String result = Protocol.topten + Protocol.separator;

	for (int i = 0; i < topTen.size(); i++)
	    result = result + ((PlayerProfile) topTen.elementAt(i)) +
		Protocol.separator;
	
	return result;
    }

    //register a logged in player. essentially adds a player to a game he 
    //will participate in. if there is no available game - make a new one
    //player can only participate in the game that matches his resolution
    public static Game register(ServerThread player, String resolution)
    {
	Game g = null;
	
	//find an emtpy game and add the player to it.
	for (int i = 0; i < games.size(); i++)
	    {	    
		g = (Game)games.elementAt(i);
		if (g.reserveSpot(resolution))
		    {
			g.addPlayer( player );
			return g;
		    }
	    }
	
	//no available games. make a new one and add the player to it
	g = new Game(game_count++, resolution);
	games.addElement(g);
	g.addPlayer( player );

	if( games.size() > maxGameCount ) 
	{
	    maxGameCount = games.size();

	    maxMGameCount = 0;
	    maxUserCount = 0;	
	    Enumeration enumGames = games.elements();
	    while(enumGames.hasMoreElements())
	    {
		g = (Game)enumGames.nextElement();
		maxUserCount += g.getHumanCount();
		if (g.getHumanCount() > 1) maxMGameCount++;
	    }
	}
	
	return g;
    }

    public static void removeGame( Game g )
    {
	games.removeElement(g);
    }

    public static void broadcast( String msg )
    {
	for (int i = 0; i < games.size(); i++)
	    ((Game)games.elementAt(i)).broadcast( msg );
    }

    //stop listening for connection
    //and disable disk access
    public void shutdown()
    {
	broadcast( Protocol.say + Protocol.separator + "Server is shuting "+
		   "down.  You may continue playing, but all profile " );
	broadcast( Protocol.say + Protocol.separator + "modifications (such "+
		   "as changing password) will be lost" );
	broadcast( Protocol.say + Protocol.separator + "If you notice any " +
		   "strange behavior in the game please report to " +
		   "alexr@scorch2000.com" );
	if ( shut_down == true )
	    {
		System.out.println("ScorchServer: shutdown already in progress");
		return;
	    }

	try
	    {
		shut_down = true;
		s_socket.close();
	    }
	catch (IOException e)
	    {
		System.out.println("ScorchServer: error shutting down " + e);
	    }
	disk.shutdown();
	serverStats.shutdown();
	//disk = null;
    }

    public void runServer()
    {
	ServerThread st = null;
	try
	    {
		//maximum of 5 connections in a queue
		s_socket = new ServerSocket(port, 5);
	    }
	catch (Exception e)
	    {
		System.err.println("SERVER: listening failed "+ e);
		System.exit(-1);
	    }

	while (!shut_down)
	    {
		//System.out.println("SERVER: accepting on port "+port);
		try
		    {
			accept = s_socket.accept();
			//accept.setSoTimeout(1000);
			
			st = new ServerThread(accept);
			
			Thread.yield();
		    }
		catch (Exception e)
		    {
			System.err.println("SERVER: listening failed "+ e);
			System.err.println("current games:\n"+
					   allGamesToString(true));
		    }
	    }

	try 
	    {
		while( games.size() > 0 )
		    Thread.sleep( 1000 );
	    }
	catch (Exception e)
	    {}
    }

    public int getRegisteredUsersCount()
    {
	return disk.getRegisteredUsersCount();
    }
}
