/*
  Class : ServerThread

  Author: Alexander Rasin

  Description: ServerThread is a separate thread that is ran for each 
  connecting client.  This thread is represents the client on the server 
  side.  It accepts messages the client sends and executes actions 
  accordingly.  Each client has access to the game it is in (once the client 
  logged in succesfully)
  ServerThread is added to the game, has a reference to it's game (for 
  broadcasts) and can ask the server when the player is attempting to log
  in, to make sure login is valid.
*/

package scorchserver;

import java.net.*;
import java.io.*;
import java.util.*;

import scorch.PlayerProfile;
import scorch.Protocol;
import scorch.utility.Debug;

public class ServerThread extends Player implements Runnable
{
    private Socket socket;
    private String hostName;

    //input and output streams
    private BufferedReader in = null;
    private PrintWriter out = null;
    private Thread self;
    private long pingtime, pongtime;
    private final static long PONGTIMEOUT = 60*1000;
    private final static long PINGTIME = 20*1000;
    private String password = null, resolution = "";
    private boolean isMaster = false;
    public Vector eotMsgs;
    
    private boolean keep_running = true, made_turn = false;

    public ServerThread(Socket socket)
    {
	this.socket = socket;
	eotMsgs = new Vector();
	pingtime = pongtime = System.currentTimeMillis();
	self = new Thread(this);
	self.start();
    }
    
    protected void makeMaster()
    {
	isMaster = true;
	//sendMessage("MASTER");
    }

    public void setReady(boolean val)
    {
        ready = val;
    }

    public boolean madeTurn()
    {
	//	System.out.println("RETURNING " + made_turn);
        return made_turn;
    }

    public void setMadeTurn( boolean set )
    {
	made_turn = set;
    }

    public boolean isMaster()
    {
	return isMaster;
    }

    //note that this method will try to resolve the host of the client
    //(unless it already has it).  So method will be used in the ScorchShell
    public String getHostName()
    {
	if (hostName == null)
	    hostName = socket.getInetAddress().getHostAddress(); 
	// bug in FreeBSD jdk1.3.1: hangs on getHostName sometimes
	//getHostName();
	
	return hostName;
    }

    //ping the client if it has been idle too long.
    //disconnect if timeout has been reached
    // return true if ok, false if timed out
    private boolean checkPing()
    {
	long now = System.currentTimeMillis();
	if (now - pongtime > PONGTIMEOUT)
	    {
		System.out.println("Player " +name+ " id "+ id +" timed out"); 
		//CHANGE THIS... PROTOCOL
		dropPlayer("your connection timed out.");
		return false;
	    }
	else 
	    {
		if(now - pingtime > PINGTIME)
		    {
			sendMessage(Protocol.ping);
			pingtime=now;
		    }
		return true;
	    }
    }

    public void sendMessage(String msg)
    {
	out.print(msg+"\r\n");
	out.flush();
    }
    
    private String getMessage()
    {
	try 
	    {
		// comment out this while loop ...
		while (keep_running && !in.ready())
		    {
			if( !checkPing() ) 
			    return null;
			try
			    {
				Thread.sleep(500);
			    }
			catch(InterruptedException e){}
		    }

		String msg = null;
		while( keep_running && msg == null )
		    {
			try
			    {
				msg = in.readLine();
			    }
			catch(OutOfMemoryError e)
			    {
				System.err.println("------------------");
				System.err.println("readLine out of memory exception");
				dropPlayer("server ran out of memory. Please report this to scorch@scorch2000.com! Please!");
				System.err.println(Debug.getMemInfo());
				System.err.println("trying to gc...");
				System.gc();
				System.err.println(Debug.getMemInfo());
				System.err.println(myGame);
				System.err.println("------------------");
			    }
			catch(SocketException e)
			    {
				return null;
			    }
			catch(InterruptedIOException e)
			    {
				checkPing();
				msg = null;
			    }
		    }

		//the thread is already disconnected
		if (!keep_running)
		    return null;
		
		//String msg = in.readLine();
		return msg;
	    }
	catch (IOException e)
	    {
		// don't print this, it's not too interesting
		//System.err.println("SERVER: reading from "+in+" failed " + e);
		dropPlayer(null);
	    }

	return null;
    }
    
    //drop a player from a game. reason == null means that the player quit
    // tuta! I've removed this synchronized because can cause deadlock
    // but doesn't do any good otherwise, I think. Instead I took out
    // seting myGame to null after we disconnect.
    public /*synchronized*/ void dropPlayer(String reason)
    {
	if (reason != null)
	    sendMessage(Protocol.disconnect + Protocol.separator + 
			"You've been disconnected because " + reason);

	// MK: this myGame != null check keeps failing
	// because other threads can call this method too
	// BS! can't synchronize on myGame because it can become null
	// just synchronize this whole method since it's the only place
	// where myGame becomes null. The question is why do we want
	// to set it to null anyway.
	//synchronized( myGame )
	//    {
		if (myGame != null)
		    {
			//actually leave the game. the player is removed first so 
			//that the broadcasts only go to other players
			myGame.leave(this);

			myGame.checkTurn();// id );
			//myGame = null;
		    }
	//    }

	disconnect();
    }

    //Not sure if this has to be synchronized, but I guess not...
    //to be tested.
    // is this synchronized causing hanging threads? We will find out.
    private /*synchronized */ void disconnect()
    {
	if (socket == null)
	    return;
	try 
	    {
		in.close();
		out.close();
		socket.close();
	    }
	catch (IOException e) { System.err.println(e); }
	socket = null;
	keep_running = false;
    }

    //see if the message received is a valid login.  The player cannot
    //join a game unless s/he provides a valid login.  This method parses
    //the input and reads the profile from server if applicable.
    private boolean logIn(String msg)
    {
	boolean loggedin = false;
	StringTokenizer st;

	password = "";

	st = new StringTokenizer(msg, ""+Protocol.separator);
	
	if (st.countTokens() < 2) 
	    return false;

	if (((msg = st.nextToken()).equals(Protocol.login)) ||
	    (msg.equals(Protocol.newplayer)))
	    {
		if (!msg.equals(Protocol.newplayer))
		    {
			if (st.hasMoreTokens()) 
			    name = st.nextToken();
			if (st.hasMoreTokens()) 
			    password = st.nextToken();
			//game resolution.
			if (st.hasMoreTokens()) 
			    resolution = st.nextToken();
		    }

		profile = ScorchServer.lookupPlayer(name);
		
		//System.out.println(" request " + msg + " found " + profile);

		if (msg.equals(Protocol.newplayer)&&(profile == null))
		    {
			/*	if (st.hasMoreTokens())
			    profile = new PlayerProfile
				(name, password, st.nextToken());
			else
			    profile = new PlayerProfile(name, password,"");
			*/
			profile = new PlayerProfile(st);
			name = profile.getName();
			password = profile.getPassword();
			while (st.hasMoreTokens()) 
			    resolution += st.nextToken();

			if (ScorchServer.lookupPlayer(name) != null)
			    profile = new PlayerProfile(null, "", "");
			else
			    //create a new player and write his profile 
			    //to disk
			    ScorchServer.newPlayer(profile);
		    }
		else 
		    //new user command failed
		    if (msg.equals(Protocol.newplayer))
			profile = new PlayerProfile(null, null);
		
		//guest login. password = desired user name
		if (name.equals(Protocol.guest))
		    {
			if (ScorchServer.lookupPlayer(password) == null)
			    {
				profile = new PlayerProfile(name, password,"");
				name = password;
				password = Protocol.guest;
			    }
			else
			    //guest 
			    profile = new PlayerProfile(null, null);
		    }
		
		if ((profile != null) && (profile.getName() == null))
		    {
			sendMessage(Protocol.loginfailed+Protocol.separator+
				    name + Protocol.separator + 
				    Protocol.usernametaken);
			loggedin = false;
		    }
		else
		if (ScorchServer.alreadyPlaying(name))
		    {
			//System.out.println(ScorchServer.alreadyPlaying(name));
			sendMessage(Protocol.loginfailed+Protocol.separator+
				    name + Protocol.separator + 
				    Protocol.alreadyloggedin);
			loggedin = false;
		    }
		else
		if (profile == null)
		    {
			sendMessage(Protocol.loginfailed+Protocol.separator+
				    name + Protocol.separator + 
				    Protocol.wrongusername);
			loggedin = false;
		    }
		else
		if ((password==null ||!password.equals(profile.getPassword())))
		    {
			sendMessage(Protocol.loginfailed+Protocol.separator+
				    name+Protocol.separator+
				    Protocol.wrongpassword); 
			loggedin = false;
		    }
		else
		    loggedin = true;
	    }
	return loggedin;
    }
    
    //initialize input/output streams for connection
    private void initStreams()
    {
	try 
	    {
		in = new BufferedReader(new InputStreamReader
					(socket.getInputStream()));
		//setting autoflush to true... wonder if it works
		out = new PrintWriter(socket.getOutputStream(), true);
	    }
	catch (IOException e)
	    {
		System.err.println("ServerThread: connection failed " + e);
		System.exit(1);
	    }
    }

    //the run method of the thread.  It continuously listends to client 
    //messages and responds/performs appropriate action.
    public void run()
    {
	String msg, command;
	StringTokenizer st;
	boolean loggedin = false;

	try
	    {
		initStreams();
	    }
	catch(OutOfMemoryError e)
	    {
		System.err.println("ServerThread.run: initStreams out of memory, will drop the player");
		disconnect();
		System.gc();
		return;
	    }
	
	do 
	    {
		//wait for a successful login command (or a quit command)
		msg = getMessage();
		
		if (msg == null || msg.equals(Protocol.quit))
		    {
			disconnect();
			return;
		    }
	    }
	while (!logIn(msg));
	
	System.out.println(" User name "+name+" logged in.");

	myGame = ScorchServer.register(this, resolution);
	checkPing();
	
	while (keep_running && ((msg = getMessage()) != null))
	    {
		pingtime=pongtime=System.currentTimeMillis(); 

		st = new StringTokenizer(msg, Protocol.separator+"");

		//no command (empty string?)
		if (st.countTokens() == 0)
		    continue;

		//the client command
		command = st.nextToken();

		//answer to ping can be ignored
		if (command.equals(Protocol.pong))
		    continue;

		//shout and setgameoptions needs to be simply broadcasted
		if (command.equals(Protocol.shout))
		    {
			//System.out.println(msg);
			myGame.broadcast(msg);
			continue;
		    }
		if (command.equals(Protocol.jvminfo))
		    {
			jvm = st.nextToken().toString();
			continue;
		    }
		//memorize game options (send them then the game starts)
		if (command.equals(Protocol.setgameoptions) && isMaster())
		    {
			myGame.gameStarted();
			myGame.setOptions(msg);

			//check if all the options neccesary have been
			//received and the clients can be initialized
			myGame.checkInitialize();
			continue;
		    }
		if (command.equals(Protocol.endofturn))
		    {
			setReady( true );//ready = true;
			//if all players are ready for the next turn,
			//notify clients of who goes next
			
			//DEBUG
			eotMsgs.addElement(msg);
			myGame.checkSync(name);

			myGame.checkTurn();// id );

			//very interesting timing bug.  if you set make_turn
			//to false before checkTurn, when if the master
			//manages to send shoot, send EOT and when MassKill,
			//the server thinks that master sent masskill _before_
			//it's turn since the variable is already reset.
			//made_turn = false; -- this did not help :)
			//made_turn variable set elsewhere (in Game.class)
			continue;
		    }
		if (command.equals(Protocol.clientlog))
		    {
			//DEBUG?
			Disk.recordClientLog(msg, id);
			if( myGame != null )
			    {
				myGame.broadcast(Protocol.shout + Protocol.separator+
			      "Client " + name + " has lost sync. It is recommended "
						 +"that you restart the game now");
				myGame.broadcast(Protocol.shout + Protocol.separator+
			      "Please report this problem to meshko@scorch2000.com");
			    }
			continue;
		    }
		//memorize the player options. will broadcast them later
		if (command.equals(Protocol.setplayeroptions))
		    {
			ploptions = msg;
			//check if all the options neccesary have been
			//received and the clients can be initialized
			myGame.checkInitialize();
			continue;
		    }
		//player used a weapon in his turn, broadcast but note the fact
		if (command.equals(Protocol.useweapon))
		    {
			made_turn = true;
			
			myGame.broadcast(msg);
			continue;
		    }
		//player says something to a player from his game.
		if (command.equals(Protocol.say))
		    {
			int plid = Integer.parseInt(st.nextToken());
			String said = st.nextToken();
			Player p = null;
			
			p = myGame.findPlayerByID(plid);
			if (p != null) 
			    {
				p.sendMessage
				    (Protocol.say +Protocol.separator + said);
				if ( id != plid )
				    sendMessage
					(Protocol.say+Protocol.separator+said);
			    }
			else
			    sendMessage(Protocol.say + Protocol.separator +
					"Message not delivered, player left the game ("+said +")");
				
			continue;
		    }
		//the user requested to change his/her profile
		if (command.equals(Protocol.changeprofile))
		    {
			//make a new profile (and cut out the command 
			//that goes in front of string msg

			profile = new PlayerProfile 
			   (msg.substring(Protocol.changeprofile.length() + 1, 
					  msg.length()));

			if (profile.getPassword().equals("*"))
			    profile.setPassword(password);
			else
			    password = profile.getPassword();

			ScorchServer.changeProfile(profile);
			continue;
		    }
		//a player died. the master client will notify us.
		if (command.equals(Protocol.playerdead))
		    {
			if (isMaster)
			    myGame.playerDied
				((Integer.parseInt(st.nextToken())));
			continue;
		    }
		//master performs a mass kill. done to finish a round
		if (command.equals(Protocol.masskill))
		    {
			if ( isMaster )
			    myGame.massKill(this);
			continue;
		    }
		//maximum players in the game
		if (command.equals(Protocol.setmaxplayers))
		    {
			if (st.hasMoreTokens())
			    myGame.setMaxPlayers
				(Integer.parseInt(st.nextToken()));
			continue;
		    }
		//add ai player
		if (command.equals(Protocol.addaiplayer))
		    {
			if (!st.hasMoreTokens())
			    sendMessage(Protocol.commandfailed);
			else
			    {
				//client checks number of AI players added
				myGame.reserveSpot();
				//add new AI player of type that was recieved
				myGame.addPlayer(new AIPlayer(st.nextToken()));
			    }
			continue;
		    }

		if (command.equals(Protocol.topten))
		    {
			sendMessage(ScorchServer.getTopTen());
			continue;
		    }
		if (command.equals(Protocol.boot))
		{
			if (!st.hasMoreTokens())
			{			
			    sendMessage(Protocol.commandfailed);
			}
			else
			{
				Player bp = myGame.findPlayerByID(Integer.parseInt(st.nextToken()));
				bp.dropPlayer("master booted you from this game");
				continue;
			}
		}		
		if (command.equals(Protocol.quit))
		    {
			dropPlayer(null);
			continue;
		    }

		//this is a temporary - unknown messages are broadcasted
		//for debug purposes.
		myGame.broadcast(msg);

		// might need this if err is redirected to file
		System.err.flush(); 
	    }

	// make sure that the socket is closed
	if( keep_running )
	    dropPlayer(null);
	    //disconnect();
    }

    public void popEOTMsg()
    {
	eotMsgs.removeElementAt(0);
    }

    public String getEOTMsg()
    {
	if( eotMsgs.size() == 0 )
	{
	    return null;
	}
	else
	{
	    String msg = (String)eotMsgs.elementAt(0);
	    //eotMsgs.removeAt(0);
	    return msg;
	}
    }
}

